﻿// ****************************************************************************
// *                                                                          *
// * © Copyright 2020 RapidApps Sweden                                        *
// *                                                                          *
// * File       : Program.cs                                                  *
// * Created    : 2020-05-08                                                  *
// * Created by : metscore                                                    *
// *                                                                          *
// ****************************************************************************

using CyUSB;
using PassMarkUSBPlugLib;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;
using static PassMarkUSBTestPlug.PassMarkDefines;

namespace PassMarkUSBTestPlug
{
    public class USBLoopBackTestFunctions
    {

        // .Net from Cypress
        // https://www.cypress.com/documentation/software-and-drivers/ez-usb-fx3-software-development-kit

        // Cypress SDK
        // https://www.cypress.com/documentation/software-and-drivers/ez-usb-fx3-software-development-kit

        // Passmark 'support'
        // https://www.passmark.com/forum/burnintest/47553-c-api-for-usb-3-0-api


        // Please note however that the current Cypress driver does not seem to be able to distinguish
        // which physical USB port that the Passmark device is connected to.
        // The property USBAddress seems to be the same disregarding what physical port device is connected to.

        private USBDeviceList usbDevices;

        private int maxBufferTransferSize;
        private Action<object, string> infoNotifier;
        private Action<object, string, Exception> errorNotifier;
        private object tag;
        private bool forceToUseUSB3 = true;
        private const string NoDeviceSelectedErrorMessage = "No device has been selected!";

        public bool AutomaticallyAddNewLine { get; set; } = false;
        public bool ForceToUseUSB3
        {
            get
            {
                return forceToUseUSB3;
            }
            set
            {
                forceToUseUSB3 = value;
                NotifyOnInformation(forceToUseUSB3 ? "Forcing to use USB3" : "USB2 or UB3 usage accepted");
            }
        }

        public USBLoopBackTestFunctions(object tag, Action<object, string> infoNotifier, Action<object, string, Exception> errorNotifier)
        {
            usbDevices = null;
            this.tag = tag;
            this.infoNotifier = infoNotifier;
            this.errorNotifier = errorNotifier;
        }

        public bool Initialize()
        {
            try
            {
                NotifyOnInformation("Initializing...");
                usbDevices = new USBDeviceList(CyConst.DEVICES_CYUSB);
                if (usbDevices is null)
                {
                    NotifyOnError("Unable to get any devices.");
                    return false;
                }
                if (usbDevices.Count == 0)
                {
                    NotifyOnError("Did not find any PassMark devices connected to the computer.");
                    return false;
                }
                NotifyOnInformation($"Found {usbDevices.Count} device(s). Will select first one found.");
                SelectedUSBDevice = (CyUSBDevice)usbDevices[0];
                return true;
            }
            catch (Exception ex)
            {
                NotifyOnError($"Failed to initialize. {ex.Message}");
                return false;
            }
        }

        private void NotifyOnInformation(string infoMessage)
        {
            infoNotifier?.Invoke(tag, $"{infoMessage}{(AutomaticallyAddNewLine ? Environment.NewLine : "")}");
        }

        private void NotifyOnError(string errorMessage, Exception ex = null)
        {
            if (errorNotifier != null)
            {
                errorNotifier(tag, errorMessage, ex);
            }
            else
            {
                throw new Exception(errorMessage, ex);
            }
        }

        public CyUSBDevice SelectedUSBDevice { get; private set; }

        public int Count
        {
            get
            {
                return usbDevices is null ? int.MinValue : usbDevices.Count;
            }
        }

        public PassMarkTestResult LastTestResult { get; private set; }
        public string Manufacturer => SelectedUSBDevice?.Manufacturer;
        public string Product => SelectedUSBDevice?.Product;
        public string FriendlyName => SelectedUSBDevice?.FriendlyName;
        public string SerialNumber => SelectedUSBDevice?.SerialNumber;
        public string DriverName => SelectedUSBDevice?.DriverName;
        public string DeviceLocation => SelectedUSBDevice?.Path;
        public int USBAddress => SelectedUSBDevice is null ? int.MinValue : SelectedUSBDevice.USBAddress;
        public string USBPID => $"0x{SelectedUSBDevice?.ProductID:X4}";
        public string USBVID => $"0x{SelectedUSBDevice?.VendorID:X4}";
        public string USBSpeed => $"{(SelectedUSBDevice is null ? "--" : SelectedUSBDevice.bHighSpeed ? "HS" : SelectedUSBDevice.bSuperSpeed ? "SS" : "FS")}";
        public bool ShowProgressTick { get; set; } = false;
        public bool ShowResultOfMeasurementInLogOnMeasurementCompletion { get; set; } = true;
        public bool PerformDataValidation { get; set; } = true;


        public void ShowConnectedDeviceInfo()
        {
            try
            {
                if (SelectedUSBDevice is null)
                {
                    NotifyOnError(NoDeviceSelectedErrorMessage);
                    return;
                }
                NotifyOnInformation($"Manufacturer : {Manufacturer}");
                NotifyOnInformation($"Product : {Product}");
                NotifyOnInformation($"Friendly name : {FriendlyName}");
                NotifyOnInformation($"Serial number : {SerialNumber}");
                NotifyOnInformation($"Driver name : {DriverName}");
                NotifyOnInformation($"Device location : {DeviceLocation}");
                NotifyOnInformation($"USB address : {USBAddress}");
                NotifyOnInformation($"VID / PID : {USBVID} / {USBPID}");
                NotifyOnInformation($"Speed : {USBSpeed}");
            }
            catch (Exception ex)
            {
                NotifyOnError($"Failed to show connected device info. {ex.Message}");
            }
        }

        public void ShowCurrentUSBVoltage(CyUSBDevice device)
        {
            try
            {
                var voltage = GetUSBVoltage(device);
                NotifyOnInformation($"Current USB voltage : {voltage:N2} V");
            }
            catch (Exception ex)
            {
                NotifyOnError($"Failed to show USB voltage. {ex.Message}");
            }
        }

        public bool ExecuteLoopbackTestForSpecificTime(int limitTestTimeInSeconds)
        {
            if (SelectedUSBDevice is null)
            {
                NotifyOnError(NoDeviceSelectedErrorMessage);
                return false;
            }
            NotifyOnInformation($"Starting test for {limitTestTimeInSeconds} seconds...");
            return ExecuteLoopbackTest(SelectedUSBDevice, limitTestTimeInSeconds * 1000, int.MinValue);
        }

        public void ShowCurrentUSBVoltage()
        {
            if (SelectedUSBDevice is null)
            {
                NotifyOnError(NoDeviceSelectedErrorMessage);
                return;
            }
            ShowCurrentUSBVoltage(SelectedUSBDevice);
        }


        public bool ExecuteLoopbackTestForSpecificNumberOfMegaBytes(int limitTestAmountInMegaBytes)
        {
            if (SelectedUSBDevice is null)
            {
                NotifyOnError(NoDeviceSelectedErrorMessage);
                return false;
            }
            NotifyOnInformation($"Starting test for {limitTestAmountInMegaBytes} MBytes...");
            var limitTestAmountInBytes = limitTestAmountInMegaBytes * 1024 * 1024;
            return ExecuteLoopbackTest(SelectedUSBDevice, int.MinValue, limitTestAmountInBytes);
        }

        private bool OkToExecuteTest(Stopwatch sw, int limitTestTime, int nbrOfSentBytes, int limitTestAmountInBytes)
        {
            if (limitTestTime != int.MaxValue && limitTestTime != int.MinValue)
            {
                return (sw.ElapsedMilliseconds < limitTestTime);
            }
            return (nbrOfSentBytes < limitTestAmountInBytes);
        }

        private long lastDisplayedTickms = long.MinValue;
        private int lastDisplayedMB = int.MinValue;
        private void ShowTick(Stopwatch sw, int limitTestTime, int nbrOfSentBytes, int limitTestAmountInBytes)
        {
            if (limitTestTime != int.MaxValue && limitTestTime != int.MinValue)
            {
                if (lastDisplayedTickms + 1000 < sw.ElapsedMilliseconds)
                {
                    lastDisplayedTickms = sw.ElapsedMilliseconds;
                    var tempHolder = AutomaticallyAddNewLine;
                    AutomaticallyAddNewLine = false;
                    NotifyOnInformation(".");
                    AutomaticallyAddNewLine = tempHolder;
                }
                return;
            }
            var numberOfMBSent = nbrOfSentBytes / 1024.0 / 1024.0;
            if (lastDisplayedMB + 5 < numberOfMBSent)
            {
                lastDisplayedMB = (int)numberOfMBSent;
                var tempHolder = AutomaticallyAddNewLine;
                AutomaticallyAddNewLine = false;
                NotifyOnInformation(".");
                AutomaticallyAddNewLine = tempHolder;
            }
        }


        private bool ExecuteLoopbackTest(CyUSBDevice device, int limitTestTime, int limitTestAmountInBytes)
        {
            try
            {
                maxBufferTransferSize = LOOPBACK_BUFFER_SIZE * LOOPBACK_BUFFER_COUNT;

                // ConfigurePlug plug does not seem to be necessary and also prevents the plug from transmitting data  :-(
                // It is being used in USB3Console.cpp but is not clear if it is only needed in that C++ application
                // due to a different underlying API. (We're using the C# wrapper implementation and there are differences
                // between the two, for instance there is no need to open the device in the C# .NET implementation)
                // ConfigurePlug(device, USB30_SPEED.USB30ST_SS);

                var systemEndPoints = GetEndPoints(new List<CypressEndPoint>(), device.Tree);
                var bulkEndPoints = GetBulkEndPoints(systemEndPoints, device, ForceToUseUSB3);

                var inEndPoint = bulkEndPoints.InEndPoint;
                var outEndPoint = bulkEndPoints.OutEndPoint;

                inEndPoint.TimeOut = 100;
                outEndPoint.TimeOut = 100;

                outEndPoint.Abort();
                outEndPoint.Reset();

                inEndPoint.Abort();
                inEndPoint.Reset();

                LastTestResult = null;
                int nbrOfSentBytes = 0;
                int nbrOfReceivedBytes = 0;
                int nbrOfWriteErrors = 0;
                int nbrOfReadErrors = 0;
                int nbrOfDataErrors = 0;

                var dataToSend = GetTransferData(maxBufferTransferSize);

                Stopwatch sw = new Stopwatch();
                sw.Start();
                while (OkToExecuteTest(sw, limitTestTime, nbrOfSentBytes, limitTestAmountInBytes))
                {
                    var (writeDataResultStatus, currentTransferSize) = WriteDataToDevice(outEndPoint, dataToSend);
                    if (!writeDataResultStatus)
                    {
                        nbrOfWriteErrors++;
                        continue;
                    }

                    nbrOfSentBytes += currentTransferSize;

                    if (currentTransferSize != maxBufferTransferSize)
                    {
                        nbrOfWriteErrors++;
                        continue;
                    }

                    if (writeDataResultStatus)
                    {
                        var (readDataResultStatus, dataReceived) = ReadDataFromDevice(inEndPoint, currentTransferSize);
                        if (!readDataResultStatus)
                        {
                            nbrOfReadErrors++;
                            continue;
                        }
                        var dataReceivedLength = dataReceived.Length;
                        nbrOfReceivedBytes += dataReceivedLength;

                        if (currentTransferSize != dataReceivedLength)
                        {
                            nbrOfReadErrors++;
                            continue;
                        }

                        sw.Stop();
                        if (ShowProgressTick) ShowTick(sw, limitTestTime, nbrOfSentBytes, limitTestAmountInBytes);
                        if (PerformDataValidation)
                        {
                            var (dataValid, nbrOfErrors) = CheckBuffersDifference(dataToSend, dataReceived, dataReceivedLength);

                            if (!dataValid)
                            {
                                nbrOfDataErrors += nbrOfErrors;
                            }
                        }
                        sw.Start();
                    }
                }
                sw.Stop();

                var success = (nbrOfWriteErrors == 0) && (nbrOfReadErrors == 0) && (nbrOfDataErrors == 0);


                LastTestResult = new PassMarkTestResult(success, sw.Elapsed, nbrOfSentBytes, nbrOfReceivedBytes, nbrOfWriteErrors, nbrOfReadErrors, nbrOfDataErrors);
                if (ShowResultOfMeasurementInLogOnMeasurementCompletion)
                {
                    NotifyOnInformation("");
                    ShowResultOfMeasurementInLog();
                }

                return success;
            }
            catch (Exception ex)
            {
                NotifyOnError($"Failed to execute loopback test. {ex.Message}");
                return false;
            }
        }

        public void ShowResultOfMeasurementInLog()
        {
            // https://physics.nist.gov/cuu/Units/binary.html
            try
            {
                NotifyOnInformation($"Test time : {LastTestResult.MeasurementTime.Minutes:D2}:{LastTestResult.MeasurementTime.Seconds:D2}.{LastTestResult.MeasurementTime.Milliseconds:D3}");
                NotifyOnInformation($"Number of bytes sent : {LastTestResult.NbrOfSentBytes} [{(int)(LastTestResult.NbrOfSentBytes / 1024.0 / 1024.0):F1} MiB] [{(int)(LastTestResult.NbrOfSentBytes / 1000.0 / 1000.0):F1} MB]");
                NotifyOnInformation($"Number of bytes received : {LastTestResult.NbrOfReceivedBytes} [{(int)(LastTestResult.NbrOfReceivedBytes / 1024.0 / 1024.0):F1} MiB] [{(int)(LastTestResult.NbrOfReceivedBytes / 1000.0 / 1000.0):F1} MB]");
                NotifyOnInformation($"Transfer speed : {(LastTestResult.NbrOfSentBytes / 1024.0 / 1024.0 / LastTestResult.MeasurementTime.TotalSeconds):F2} MB/s [{(LastTestResult.NbrOfSentBytes * 8.0 / 1024.0 / 1024.0 / LastTestResult.MeasurementTime.TotalSeconds):F2} Mibit/s] [{(LastTestResult.NbrOfSentBytes * 8.0 / 1000.0 / 1000.0 / LastTestResult.MeasurementTime.TotalSeconds):F2} Mbit/s]");
                NotifyOnInformation($"Number of write errors : {LastTestResult.NbrOfWriteErrors}");
                NotifyOnInformation($"Number of read errors : {LastTestResult.NbrOfReadErrors}");
                NotifyOnInformation($"Number of data errors : {(PerformDataValidation ? LastTestResult.NbrOfDataErrors.ToString() : "No data validation performed")}");
                NotifyOnInformation($"Success : {(LastTestResult.Success ? "No errors" : "There were errors !!")}");
            }
            catch (Exception ex)
            {
                NotifyOnError($"Failed to show measurement in log. {ex.Message}");
            }
        }

        private (bool Result, int bytesWritten) WriteDataToDevice(CyBulkEndPoint outEndPoint, byte[] dataToSend)
        {
            var packetMode = true;
            var currentTransferSize = dataToSend.Length;
            var writeDataResultStatus = outEndPoint.XferData(ref dataToSend, ref currentTransferSize, packetMode);
            if (!writeDataResultStatus)
            {
                outEndPoint.Abort();
                outEndPoint.Reset();
            }
            return (writeDataResultStatus, currentTransferSize);
        }

        private (bool Result, byte[] DataRead) ReadDataFromDevice(CyBulkEndPoint inEndPoint, int nbrOfBytesToRead)
        {
            byte[] dataToReceive = new byte[nbrOfBytesToRead];
            var dataReceiveLength = dataToReceive.Length;

            var readDataResultStatus = inEndPoint.XferData(ref dataToReceive, ref dataReceiveLength);

            return (readDataResultStatus, dataToReceive);

        }

        private (bool Equal, int NbrOfFaults) CheckBuffersDifference(byte[] buffer1, byte[] buffer2, int lengthToCheck)
        {
            if (buffer1 is null || buffer2 is null) throw new Exception("Buffers can not be null");
            if (buffer1.Length < lengthToCheck || buffer2.Length < lengthToCheck) throw new Exception("Buffer is smaller than requested length check");
            if (buffer1.Length.Equals(buffer2.Length))
            {
                ReadOnlySpan<byte> B1 = buffer1;
                ReadOnlySpan<byte> B2 = buffer2;
                var result = B1.SequenceEqual(B2);
                if (result) return (true, 0);
                int nbrOfErrorsFound = 0;
                for (int i = 0; i < buffer1.Length; i++)
                {
                    nbrOfErrorsFound += buffer1[i].Equals(buffer2[i]) ? 0 : 1;
                }
                return (false, nbrOfErrorsFound);

            }
            ReadOnlySpan<byte> partialB1 = MemoryExtensions.AsSpan(buffer1, 0, lengthToCheck);
            ReadOnlySpan<byte> partialB2 = MemoryExtensions.AsSpan(buffer2, 0, lengthToCheck);
            var partialResult = partialB1.SequenceEqual(partialB2);
            if (partialResult) return (partialResult, 0);
            int nbrOfErrorsFoundInPartial = 0;
            for (int i = 0; i < lengthToCheck; i++)
            {
                nbrOfErrorsFoundInPartial += buffer1[i].Equals(buffer2[i]) ? 0 : 1;
            }
            return (false, nbrOfErrorsFoundInPartial);
        }

        private void ConfigurePlug(CyUSBDevice device, USB30_SPEED speed)
        {
            PassMarkConfig usb3Config = new PassMarkConfig();
            usb3Config.dma_mode = (byte)USB_TEST_TYPE.DMA_LOOPBACK_AUTO;

            usb3Config.ep_type = (byte)USB_EP_TYPE.USB_EP_BULK;
            usb3Config.ep_in = 1;
            usb3Config.ep_out = 1;
            usb3Config.ss_burst_len = 1;
            usb3Config.polling_interval = 1;
            usb3Config.hs_bulk_nak_interval = 0;
            usb3Config.iso_bytes_per_bus_interval = 0;
            usb3Config.iso_transactions_per_bus_interval = 1;

            usb3Config.speed = (byte)speed;
            usb3Config.buffer_count = LOOPBACK_BUFFER_COUNT;
            usb3Config.buffer_size = LOOPBACK_BUFFER_SIZE;

            WriteVendorCommand(device, SET_CONFIG, GetBytesForConfig(usb3Config));

            // Plug will now enumerate
            NotifyOnInformation("Plug will now enumerate....");
            Thread.Sleep(3000);
            if (!Initialize())
            {
                NotifyOnError("Failed to initialize plug after enumeration");
                throw new Exception("Failure to re-initialize after enumeration");
            }
            SelectedUSBDevice.ReConnect();
        }

        private byte[] GetBytesForConfig(PassMarkConfig buf)
        {
            int dataLength = Marshal.SizeOf(buf);
            IntPtr ptr = Marshal.AllocHGlobal(dataLength);
            Marshal.StructureToPtr(buf, ptr, true);
            byte[] data = new byte[dataLength];
            Marshal.Copy(ptr, data, 0, dataLength);
            Marshal.FreeHGlobal(ptr);
            return data;
        }

        public double GetUSBVoltage()
        {
            if (SelectedUSBDevice is null)
            {
                NotifyOnError(NoDeviceSelectedErrorMessage);
                return double.NaN;
            }
            return GetUSBVoltage(SelectedUSBDevice);
        }

        public double GetUSBVoltage(CyUSBDevice device)
        {
            try
            {
                var data = ReadVendorCommand(device, GET_VOLTAGE, sizeof(double));
                if (data is null) throw new Exception("Failed to read voltage");
                var voltage = BitConverter.ToDouble(data, 0);
                return voltage;
            }
            catch (Exception ex)
            {
                NotifyOnError($"Failed to get USB voltage. {ex.Message}");
                throw;
            }
        }

        private byte[] ReadVendorCommand(CyUSBDevice device, ushort wValue, int lengthToRead)
        {
            var controlEndPoint = device.ControlEndPt;
            if (controlEndPoint is null) throw new Exception("Endpoint is not available");

            controlEndPoint.Target = CyConst.TGT_DEVICE;
            controlEndPoint.ReqType = CyConst.REQ_VENDOR;
            controlEndPoint.ReqCode = 0x00;
            controlEndPoint.Value = wValue;
            controlEndPoint.Index = 0;

            byte[] data = new byte[lengthToRead];
            var result = controlEndPoint.Read(ref data, ref lengthToRead);

            return result ? data : null;
        }

        private bool WriteVendorCommand(CyUSBDevice device, ushort wValue, byte[] data)
        {
            var controlEndPoint = device.ControlEndPt;

            controlEndPoint.Target = CyConst.TGT_DEVICE;
            controlEndPoint.ReqType = CyConst.REQ_VENDOR;
            controlEndPoint.ReqCode = 0x00;
            controlEndPoint.Value = wValue;
            controlEndPoint.Index = 0;

            var dataLength = data.Length;

            var result = controlEndPoint.Write(ref data, ref dataLength);
            return result;
        }

        private byte[] GetTransferData(int size)
        {
            byte[] outData = new byte[size];

            Random random = new Random(DateTime.Now.Millisecond);
            random.NextBytes(outData);
            return outData;
        }

        private USB30_SPEED GetUSBSpeed(CyUSBDevice usbDevice)
        {
            if (usbDevice.bHighSpeed)
            {
                return USB30_SPEED.USB30ST_HS;
            }
            else if (usbDevice.bSuperSpeed)
            {
                return USB30_SPEED.USB30ST_SS;
            }
            return USB30_SPEED.USB30ST_FS;
        }

        private (CyBulkEndPoint InEndPoint, CyBulkEndPoint OutEndPoint) GetBulkEndPoints(List<CypressEndPoint> systemIfcs, CyUSBDevice usbDevice, bool forceUSB3)
        {
            try
            {
                var outEP = systemIfcs.Find(ep => ep.EndPointType == EndPointType.Out && ep.Endpoint.Address != 0x07);
                if (outEP is null) throw new Exception($"Unable to find OUT endpoint");

                if (outEP.Endpoint.MaxPktSize < 0x4000 && forceUSB3) throw new Exception($"Out endpoint does not seem to be a USB3 endpoint");

                var inEP = systemIfcs.Find(ep => ep.EndPointType == EndPointType.In && ((ep.Endpoint.Address) & 0x7F) != 0x07);
                if (inEP is null) throw new Exception($"Unable to find matching IN endpoint");

                var inBulkEndPoint = usbDevice.EndPointOf(inEP.Endpoint.Address) as CyBulkEndPoint;
                var outBulkEndPoint = usbDevice.EndPointOf(outEP.Endpoint.Address) as CyBulkEndPoint;

                return (inBulkEndPoint, outBulkEndPoint);
            }
            catch (Exception ex)
            {
                throw;
            }
        }

        private List<CypressEndPoint> GetEndPoints(List<CypressEndPoint> result, TreeNode devTree)
        {
            const string BULK_IN_NOTIFIER = "Bulk in";
            const string BULK_OUT_NOTIFIER = "Bulk out";
            foreach (TreeNode node in devTree.Nodes)
            {
                if (node.Nodes.Count > 0)
                {
                    result = GetEndPoints(result, node);
                }
                else
                {
                    CyUSBEndPoint ept = node.Tag as CyUSBEndPoint;
                    if (ept is null)
                    {
                        // Empty
                    }
                    else if (node.Text.Contains(BULK_IN_NOTIFIER) || node.Text.Contains(BULK_OUT_NOTIFIER))
                    {
                        EndPointType type = node.Text.Contains(BULK_IN_NOTIFIER) ? EndPointType.In : node.Text.Contains(BULK_OUT_NOTIFIER) ? EndPointType.Out : EndPointType.Unknown;
                        CyUSBInterface ifc = node.Parent.Tag as CyUSBInterface;
                        var endpoint = ifc.EndPoints.ToList().Find(ep => node.Text.Contains(ep.Address.ToString("X2")));
                        result.Add(new CypressEndPoint(type, ifc.bAlternateSetting, ept.MaxPktSize, endpoint));
                    }
                }
            }
            return result;
        }
    }
}
